"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

# Module HSM_WFN
"""A module that consists of the methods used to construct the numerical wavefront
aberration from gradients.

"""

from numpy import *
from scipy.sparse import *
from scipy.interpolate import griddata, RectBivariateSpline
from scipy.sparse.linalg import lsmr, lsqr
import gc


def setup_grid(limits, gspacing):
    """Set up square mesh grids for wavefront construction.

    :param limits: the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :Returns: *ndarrays* of x and y grids
    :rtype: *tuple*

    """

    x1, x2 = limits["x"]
    y1, y2 = limits["y"]

    xr = arange(x1, x2 + gspacing, gspacing)
    yr = arange(y1, y2 + gspacing, gspacing)

    if xr[-1] > x2:
        xr = xr[:-1]

    if yr[-1] > y2:
        yr = yr[:-1]

    return meshgrid(xr, yr)


def calculate_wf(cents, grads, limits, gspacing, method="lsmr"):
    """Calculate the two-dimensional wavefront aberration array from the gradients.

    Note that ``cents`` and ``grads`` must be of the same shape.

    :param cents: the centroids to be used as the positions of the gradients
    :type cents: *ndarray*
    :param grads: the gradients from which the wavefront aberration is to be calcualted
    :type grads: *ndarray*
    :param limits: the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :param method: a method to solve the matrix equation (should be 'lsqr'
                   or 'lsmr')
    :type method: *string*
    :Uses: :func:`~HWS.HSM_WFN.interpolate_gradients`, :func:`~HWS.HSM_WFN.integrate_gradients`
    :Returns: the calculated wavefront
    :rtype: *ndarray*

    """
    gradf_x, gradf_y = interpolate_gradients(cents, grads, limits, gspacing)
    wf_t = -1.0 * integrate_gradients(
        gradf_x, gradf_y, gspacing, gspacing, method=method
    )
    wf = wf_t - wf_t.min()

    return wf


def calculate_wf_from_hsgradients(
    hsg, limits=None, gspacing=None, method="lsmr", to_interpolate=True
):
    """Calculate the two-dimensional wavefront aberration array from the an instance
    of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`.

    This is a convenience method that calculates a numerical wavefront from an
    instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>` and a set of
    other parameters as in :func:`~HWS.HSM_WFN.calculate_wf`.

    As long as a user supply an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`, the method will calculate the wavefront array from them using preset values for the other parameters.

    :param hsg: an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`,
                with ``gradients`` and ``pixel_size`` appropriately set.
    :type hsg: an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`
    :param limits: (optional) the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: (optional) the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :param method: (optional) a method to solve the matrix equation (should be 'lsqr'
                   or 'lsmr')
    :type method: *string*
    :Uses: :func:`~HWS.HSM_WFN.setup_grid`,
           :func:`~HWS.HSM_WFN.interpolate_wf_default`,
           :func:`~HWS.HSM_WFN.calculate_wf`
    :Returns: the calculated wavefront
    :rtype: *ndarray*

    """
    cents = hsg.gradients[:, 2:]
    grads = hsg.gradients[:, :2]

    if gspacing is None:
        gspacing = 4 * hsg.pixel_size

    if limits is None:
        limits = {}
        limits["x"] = array(
            [cents[:, 0].min() - gspacing, cents[:, 0].max() + gspacing]
        )

        limits["y"] = array(
            [cents[:, 1].min() - gspacing, cents[:, 1].max() + gspacing]
        )

    # parameters = {}
    # parameters['limits'] = limits
    # parameters['gspacing'] = gspacing
    # parameters['coords'] =
    if to_interpolate is True:
        xg, yg = setup_grid(limits, gspacing)
        xc, yc = xg[0, :], yg[:, 0]
        return interpolate_wf_default(
            xc, yc, calculate_wf(cents, grads, limits, gspacing)
        )
    else:
        return calculate_wf(cents, grads, limits, gspacing)


def interpolate_gradients(cents, grads, limits, gspacing, method="linear"):
    """Interpolate the gradients.

    This method calls ``griddata`` method in a module ``scipy.interpolate``.

    The input parameter ``method`` must be ``nearest``, ``linear`` or ``cubic``.

    :param cents: the centroids to be used as the positions of the gradients
    :type cents: *ndarray*
    :param grads: the gradients
    :type grads: *ndarray*
    :param limits: the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :param method: a method used to interpolate the gradients
    :type method: *string*
    :Uses: :func:`~HWS.HSM_WFN.setup_grid`
    :Returns: *ndarrays* of x and y components of the interpolated gradients
    :rtype: *tuple*

    """
    mymethod = method
    xg, yg = setup_grid(limits, gspacing)

    gradf_x = griddata(cents, grads[:, 0], (xg, yg), method=mymethod)
    gradf_y = griddata(cents, grads[:, 1], (xg, yg), method=mymethod)

    gradf_x[isnan(gradf_x)] = 0
    gradf_y[isnan(gradf_y)] = 0

    return (gradf_x, gradf_y)


def calculate_wf_prism(limits, gspacing, prism, alpha):
    """Calculate two-dimensional prism wavefront aberration array.

    This method generates a prism wavefront aberration of the magnitude ``prism``
    and the angle ``alpha``.

    :param limits: the minimum and maximum values of x and y coordinates
    :type limits: *dict*
    :param gspacing: the spacing between two neighboring points on the grid
    :type gspacing: positive *float*
    :param prism: the magnitude of the prism
    :type prism: positive *float*
    :param alpha: the angle of the prism
    :type alpha: *float*
    :Uses: :func:`~HWS.HSM_WFN.setup_grid`
    :Returns: the constructed prism wavefront array
    :rtype: *ndarray*

    """
    xg, yg = setup_grid(limits, gspacing)
    wf = prism * (xg * cos(alpha) + yg * sin(alpha))
    return wf


def interpolate_wf_default(
    xc, yc, wf, xlimits=(-511.5, 511.5), ylimits=(-511.5, 511.5), fspacing=12e-6
):
    """Interpolate a wavefront aberration array with some default parameters.

    This method can be used to produce a two-dimensional wavefront array from an
    array constructed by :func:`~HWS.HSM_WFN.integrate_gradients`.

    The resulting interpolated wavefront is of a higher resolution (1024 by 1024)
    than the integrated wavefront and thus may be more suitable for a visual
    inspection.

    :param xc: the x coordinates of the wavefront to be interpolated
    :type xc:
    :param yc: the y coordinates of the wavefront to be interpolated
    :type yc:
    :param wf: the wavefront to be interpolated
    :type wf:
    :Uses: :func:`~HS_WFP.setup_grid`
    :Returns: the interpolated wavefron array
    :rtype: *ndarray*

    """
    if type(xlimits) is not ndarray:
        xl = array(xlimits)
    else:
        xl = xlimits

    if type(ylimits) is not ndarray:
        yl = array(ylimits)
    else:
        yl = ylimits

    flimits = {}
    flimits["x"] = xl * fspacing
    flimits["y"] = yl * fspacing

    xgf, ygf = setup_grid(flimits, fspacing)

    rbs = RectBivariateSpline(xc, yc, wf.T)
    wf_intp = rbs.ev(xgf, ygf)
    # wf_intp = griddata(coords,wf.flatten(),(xgf,ygf),method = mymethod)

    # if mymethod != 'nearest':
    #     wf_nn = griddata(coords,wf.flatten(),(xg,yg),method = 'nearest')
    #     wf_bi = isnan(wf_intp)
    #     wf_intp[wf_bi] = wf_nn[wf_bi]

    return wf_intp


def integrate_gradients(gxi, gyi, dx=1, dy=1, detail=False, method="lsmr"):
    """Numerically integrate the gradients to produce a two-dimensional wavefront
    aberration array.

    :param gxi: the x components of the gradients to be integrated
    :type gxi: *ndarray*
    :param gyi: the y components of the gradients to be integrated
    :type gyi: *ndarray*
    :param dx: the spacing between two nearest points in x direction
    :type dx: positive *float*
    :param dy: the spacing between two nearest points in y direction
    :type dy: positive *float*
    :param detail: a flag to indicate whether to output the integrated wavefront only or
                   to output more information as well
    :param method: the method used to solve the matrix equation (should be either 'lsqr'
                   or 'lsmr')
    :Returns: the integrated wavefront array, (optional) the information on the calculation
    :rtype: *ndarray* or *tuple* of *ndarray* and *dict*

    """
    #  original Matlab code is from:
    #
    #  http://www.mathworks.com/matlabcentral/fileexchange/9734
    #  by John D'Errico
    #  23 Jan 2006 (Updated 27 Jan 2006)
    #  Code covered by the BSD License
    #
    #  Converted to Python by Won Kim
    gx = gxi.copy()
    gy = gyi.copy()
    (ny, nx) = shape(gx)
    # while developing
    # dx = 1
    # dy = 1

    dx = tile(dx, [nx - 1, 1])
    dy = tile(dy, [ny - 1, 1])
    rhs = zeros(2 * nx * ny)
    Af = zeros([2 * nx * ny, 6])
    L = 0

    indx = 0
    indy = arange(ny)
    ind = indy + indx * ny

    rind = tile(L + arange(ny), [2, 1]).T
    cind = array([ind, ind + ny]).T
    dfdx = tile(array([-1.0, 1.0]) / dx[0], [ny, 1])
    Af[L + arange(ny), :] = concatenate((rind, cind, dfdx), axis=1)
    rhs[L + arange(ny)] = gx[:, 0]
    L = L + ny

    if nx > 2:
        indx, indy = meshgrid(arange(nx - 2) + 1, arange(ny))
        indx = indx.T.flatten()
        indy = indy.T.flatten()
        ind = indy + indx * ny
        m = ny * (nx - 2)

        rind = tile(L + arange(m), [2, 1]).T
        cind = array([ind - ny, ind + ny]).T

        dfdx = 1.0 / (dx[indx - 1] + dx[indx])
        dfdx = dfdx * array([-1, 1])

        Af[L + arange(m), :] = concatenate((rind, cind, dfdx), axis=1)
        rhs[L + arange(m)] = gx.T.flatten()[ind]

        L = L + m

    indx = nx - 1
    indy = arange(ny)
    ind = indy + indx * ny
    rind = tile(L + arange(ny), [2, 1]).T
    cind = array([ind - ny, ind]).T
    dfdx = tile(array([-1.0, 1.0]) / dx[-1], [ny, 1])
    Af[L + arange(ny), :] = concatenate((rind, cind, dfdx), axis=1)
    rhs[L + arange(ny)] = gx[:, -1]
    L = L + ny

    indx = arange(nx)
    indy = 0
    ind = indy + indx * ny
    rind = tile(L + arange(nx), [2, 1]).T
    cind = array([ind, ind + 1]).T
    dfdy = tile(array([-1, 1]) / dy[0], [nx, 1])
    Af[L + arange(nx), :] = concatenate((rind, cind, dfdy), axis=1)
    rhs[L + arange(nx)] = gy[1, :]
    L = L + nx

    if ny > 2:
        indx, indy = meshgrid(arange(nx), arange(ny - 2) + 1)
        indx = indx.flatten()
        indy = indy.flatten()
        ind = indy + indx * ny
        m = nx * (ny - 2)

        rind = tile(L + arange(m), [2, 1]).T
        cind = array([ind - 1, ind + 1]).T

        dfdy = 1.0 / (dy[indy - 1] + dy[indy])
        dfdy = dfdy * array([-1, 1])

        Af[L + arange(m), :] = concatenate((rind, cind, dfdy), axis=1)
        rhs[L + arange(m)] = gy.T.flatten()[ind]

        L = L + m

    indx = arange(nx)
    indy = ny - 1
    ind = indy + indx * ny
    rind = tile(L + arange(nx), [2, 1]).T
    cind = array([ind - 1, ind]).T
    dfdy = tile(array([-1, 1]) / dy[-1], [nx, 1])
    Af[L + arange(nx), :] = concatenate((rind, cind, dfdy), axis=1)
    rhs[L + arange(nx)] = gy[-1, :]

    ii = Af[:, 0:2].flatten()
    jj = Af[:, 2:4].flatten()
    vv = Af[:, 4:6].flatten()

    A = csc_matrix((vv, (ii, jj)), shape=(2 * nx * ny, nx * ny))
    # A = csc_matrix((Af[:,4:6].T,(Af[:,0:2].T,Af[:,2:4].T)),shape = (2*nx*ny,nx*ny))

    # gc.collect()
    # this is where rhs shape changes
    # A0 = A[:,0].copy()
    # dcterm = A0*dc
    # gc.collect()
    # rhs = rhs - dcterm
    # gc.collect()
    # rhs = rhs - dcterm.toarray()
    # rhs = rhs - A.toarray()[:,0]*dc
    # fhat = lsmr(A[:,1:],rhs,atol=0,btol=0,conlim=0)
    gc.collect()
    if method == "lsqr":
        result = lsqr(A[:, 1:], rhs)
    else:
        result = lsmr(A[:, 1:], rhs)

    fhat = result[0]
    wf = concatenate((array([0]), fhat)).reshape(ny, nx, order="F")

    if detail == True:
        resdat = {}
        if method == "lsmr":
            resdat["method"] = "lsmr"
            resdat["x"] = result[0]
            resdat["A"] = A
            resdat["b"] = rhs
            resdat["stop_reason"] = result[1]
            resdat["no_iteration"] = result[2]
            resdat["normr"] = result[3]
            resdat["normar"] = result[4]
            resdat["norama"] = result[5]
            resdat["conda"] = result[6]
            resdat["normx"] = result[7]
        else:
            resdat["method"] = "lsqr"
            resdat["x"] = result[0]
            resdat["A"] = A
            resdat["b"] = rhs
            resdat["stop_reason"] = result[1]
            resdat["no_iteration"] = result[2]
            resdat["r1norm"] = result[3]
            resdat["r2norm"] = result[4]
            resdat["anorm"] = result[5]
            resdat["acond"] = result[6]
            resdat["arnorm"] = result[7]
            resdat["xnorm"] = result[8]
        return wf, resdat
    else:
        return wf



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *
from HWS.HS_Gradients import *
from scipy import linalg


class HS_WFP:
    """A class used to fit a sum of two-dimensional polynomials to the gradients to
    characterise the transverse wavefront abberation.

    The class performs a least squares fit of the gradients to a sum of the polynomials
    (the instance variables ``X_bases`` and ``Y_bases``)
    and calculates the polynomial coeefficients. Once the coefficients are obtained, a
    user can also construct the corresponding transverse wavefron aberration map using
    the coefficients and the functions in ``poly_bases``.

    The polynomial coefficients can also be used to (in a limited sense) calculate
    the Zernike (up to the second order) and aome Seidel coefficients.

    **Instantiation**

    Upon instantiation, the class will build the basis functions (or bases) and
    determine how many basis functions are required for the operation.

    Two parameters a user can choose are ``hsgr`` and ``order``, ``hsgr`` is an instance
    of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>` and ``order`` is
    an integer from 1 to 6.

    ``order`` specifies the maximum power of polynomials that the class will use to
    characterise the wavefront. The class allows the maximum order of 6, which is
    the default value of ``order``.

    :func:`~HS_WFP.determine_no_of_bases` which is run during instantiation, sets
    ``no_of_bases`` to a value consistent with ``order``.

    Two variables ``gradients`` and ``centroids`` are generated from ``hsgradients``
    if it is set as an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`
    by using the input parameter ``hsg``.

    A user should note that the centroids (i.e., the 3rd and the 4th columns of the
    ``gradients`` instance variable of ``hsgradients``) should be in metre units about
    a suitably chosen origin (normally the centre of an image), which should automatically
    the case when the gradients are constructed using
    :func:`HS_Gradients.construct_gradients() <HWS.HS_Gradients.HS_Gradients.construct_gradients>` as normal.

    **Basis Functions**

    There are three sets of basis functions in this class:

    - polynomial basis functions (``poly_all_bases``) are the polynomial basis functions
      that can be used in the polynomial wavefront reconstruction using the method
      :func:`~HS_WFP.calculate_wf`. These functions are NOT used in least squares fitting.

    - X basis functions (``X_all_bases``) are the partial x derivatives of the polynomial
      basis functions, and are used to fit the x components of the gradients.

    - Y basis functions (``Y_all_bases``) are the partial y derivatives of the polynomial
      basis functions, and are used to fit the y components of the gradients.

    **Instance Variables**

    - ``unitization_length``: the length factor used to scale the coordinates of the
      centroids

      The variable is relevant only when ``to_unitize`` is ``True``.

      :Type: positive *float*
      :Required by: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.deunitize`

    - ``to_unitize``: a *boolean* to indicate whether the centroids should be scaled or not

      If ``True``, the lengths in centroids and basis fucntions will be scaled by
      ``unitization_length``.

      :Type: *boolean*
      :Requried by: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`,
                    :func:`~HS_WFP.calculate_gradients`,

    - ``order``: the maximum power of the polynomials to be used in the least squares fit of
      the gradients

      The default value is 6, which is also the maximum value that is currently available.

      :Type: *integer* (from 1 to 6)
      :Required by: :func:`~HS_WFP.determine_no_of_bases`

    - ``no_of_bases``: the number of polynomial basis functions to be used in the least
      squares fit

      A valid value of this variable must be one of those in the class variable
      ``no_of_bases_array``, depending on ``order``. For example if ``order`` is
      3, the valid ``no_of_bases`` is 9.

      With 6 being the default value of ``order``, the valid ``no_of_bases`` is 27.

      :Type: *integer* (from the class variable ``no_of_bases_array``)
      :Required by: :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`,
                    :func:`~HS_WFP.deunitize`,
                    :func:`~HS_WFP.build_bases`
      :Updated by: :func:`~HS_WFP.determine_no_of_bases`

    - ``hsgradients``: an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`

      During instantiation, it is split into ``gradients`` and ``centroids`` which are
      subsequently used by other methods in the class.

      :Type: an instance of :class:`HS_Gradients <HWS.HS_Gradients.HS_Gradients>`

    - ``centroids``: the centroids to be used in the least squares fit

      During instantiation, the third and fourth columns of ``hsgradients.gradients`` are
      assigned to this variable.

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.build_XY`,
                    :func:`~HS_WFP.build_XY_nu`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``gradients``: the gradients to be used in the least squares fit

      During instantiation, the first and second columns of ``hsgradients.gradients`` are
      assigned to this variable.

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`,
                    :func:`~HS_WFP.compute_SSR`

    - ``poly_all_bases_names``: the names of all polynomial basis functions that can be
      used in the polynomial wavefront reconstruction

      This variable contains all of the polynomial basis function names that can be used
      in the least squares fitting of the gradients.

      The naming format is xmyn where m, n are positive integers: for eample `x2y3` refers
      to a function (x^2)*(y^3).

      :Type: *list* of *strings*
      :Required by: :func:`~HS_WFP.build_bases`
      :Updated by: :func:`~HS_WFP.build_all_bases`

    - ``poly_bases_names``: the names of the polynomial basis functions used in the
      polynomial wavefront reconstruction

      This variable is derived from ``poly_all_bases_names`` by taking a number of elements
      (set by ``no_of_bases``) starting from the first in ``poly_all_bases_names``.

      :Type: *list* of *strings*
      :Required by: :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`
      :Updated by: :func:`~HS_WFP.build_bases`

    - ``poly_all_bases``: all polynomial basis functions that can be used in the polynomial
      wavefront reconstruction

      :Type: *list*
      :Required by: :func:`~HS_WFP.calcualte_wf`
      :Updated by: :func:`~HS_WFP.build_all_bases`

    - ``poly_coeffs``: a *dict* of the coefficients of the polynomial basis functions
      calculated from the least squares fit

      :Type: *dict*
      :Required by: :func:`~HS_WFP.calcualte_wf`
      :Updated by: :func:`~HS_WFP.compute_poly_coeffs`,
                   :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``poly_coeffs_array``: an *ndarray* of the coefficients of the polynomial bases
      functions calculated from the least square fit

      This is an *ndarray* version of ``poly_coeffs``, a convenience

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.calculate_gradients`,
                    :func:`~HS_WFP.calculate_wf`,
                    (optional) :func:`~HS_WFP.compute_seidel_coeffs`,
                    (optional) :func:`~HS_WFP.compute_zernike_coeffs`
      :Updated by: :func:`~HS_WFP.compute_poly_coeffs`,
                   :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``poly_coeffs_u``: a *dict* of the coefficients of the unitized polynomial basis
      functions calculated from the least squares fit

      This variable is used only when ``to_unitize`` is ``True``.

      :Type: *dict*
      :Updated by: :func:`~HS_WFP.compute_poly_coeffs`,
                   :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``poly_coeffs_array_u``: an *ndarray* of the coefficients of the unitized
      polynomial basis functions calculated from the least squares fit

      This variable is used only when ``to_unitize`` is ``True``.

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.deunitize`,
                    (optional) :func:`~HS_WFP.calculate_gradients`
      :Updated by: (optional) :func:`~HS_WFP.compute_poly_coeffs`,
                   (optional) :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``SSR``: squared sum of the residues after fitting the gradients to the polynomial
      basis functions

      :Type: positive *float*
      :Updated by: :func:`~HS_WFP.compute_SSR`

    - ``seidel_coeffs``: a *dict* of the Seidel coefficients derived from the polynomial
      basis function coefficients

      :Type: *dict*
      :Update by: (optional) :func:`~HS_WFP.compute_seidel_coeffs`

    - ``zernike_coeffs``: an *ndarray* of the Zernike coefficients derived from the
      polynomial basis function coefficients

      :Type: *dict*
      :Update by: (optional) :func:`~HS_WFP.compute_zernike_coeffs`

    - ``X_all_bases``: a *list* of all basis functions that can be used in the least squares
      fitting of the x components of the gradients

      :Type: *list*
      :Required by: :func:`~HS_WFP.build_bases`
      :Updated by: :func:`~HS_WFP.build_all_bases`

    - ``Y_all_bases``: a *list* of all basis functions that can be used in the least squares
      fitting of the y components of the gradients

      :Type: *list*
      :Required by: :func:`~HS_WFP.build_bases`
      :Updated by: :func:`~HS_WFP.build_all_bases`

    - ``X_bases``: a *list* of all basis functions to be used in the least squares fitting
      of the x components of the gradients

      :Type: *list*
      :Required by: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.build_XY_nu`
      :Updated by: :func:`~HS_WFP.build_bases`

    - ``Y_bases``: a *list* of all basis functions to be used in the least squares fitting
      of the y components of the gradients

      :Type: *list*
      :Required by: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.build_XY_nu`
      :Updated by: :func:`~HS_WFP.build_bases`

    - ``X``: an *ndarray* of the values of the functions in ``X_bases`` evaluated at each
      coord of ``centroids``

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.build_M_matrix`, :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`,
      :Updated by: :func:`~HS_WFP.build_XY`

    - ``Y``: an *ndarray* of the values of the functions in ``Y_bases`` evaluated at each
      coord of ``centroids``

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.build_M_matrix`, :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`,
      :Updated by: :func:`~HS_WFP.build_XY`

    - ``M``: the M matrix used in the least squares fitting

      :Type: *ndarray*
      :Required by: :func:`~HS_WFP.compute_poly_coeffs`,
                    :func:`~HS_WFP.compute_poly_coeffs_weighted`
      :Updated by: :func:`~HS_WFP.build_M_matrix`

    - ``b``: the b vector used in the least squares fitting

      :Type: *ndarray*
      :Updated by: :func:`~HS_WFP.compute_poly_coeffs`,
                   :func:`~HS_WFP.compute_poly_coeffs_weighted`

    - ``weights``: an *ndarray* of weights used in the weighted least squares fitting

      This variable is used to weight (or scale) ``gradients``, ``X`` and ``Y`` before
      calculating the polynomial coefficients for the wavefront.

      As long as its shape is the same as that of ``gradients``, ``weights`` can be an
      array of any positive numbers (cannot be zero). However, in order for the weighted
      least squares fit to be meaningful, ``weights`` should be roughly proportional to
      the uncertainties of the measured gradients.

      For example, the standard deviation of each coordinate of each centroid, calculated
      using multiple image frames (or averaged images), can be used as ``weights``.

      :Type: *ndarray*
      :Updated by: :func:`~HS_WFP.compute_poly_coeffs_weighted`

    """

    no_of_bases_array = [2, 5, 9, 14, 20, 27]
    """A class variable specifying the numbers of basis functions for each order.

    For example, if the order of the polynomials is 1, the number of basis
    functions is 2 (although order 1 is hardly useful). The maximum order
    implemented in this class is six, for which we have 27 basis functions.

    :Type: *list*
    :Required by: :func:`~HS_WFP.determine_no_of_bases`

    """

    def __init__(self, hsgr=None, order=6):
        self.unitization_length = 1.0
        self.to_unitize = False
        self.order = order
        self.no_of_bases = None
        self.hsgradients = hsgr
        self.centroids = None
        self.gradients = None
        self.poly_bases_names = None
        self.poly_all_bases_names = None
        self.poly_all_bases = None
        self.poly_coeffs = None
        self.poly_coeffs_array = None
        self.poly_coeffs_u = None
        self.poly_coeffs_array_u = None
        self.SSR = None
        self.seidel_coeffs = None
        self.zernike_coeffs = None
        self.M = None
        self.b = None
        self.X = None
        self.Y = None
        self.X_bases = None
        self.Y_bases = None
        self.X_all_bases = None
        self.Y_all_bases = None
        self.weights = None

        self.build_all_bases()
        self.determine_no_of_bases(order)

        if hsgr.__class__.__name__ == "HS_Gradients":
            self.hsgradients = hsgr
            self.gradients = hsgr.gradients[:, 0:2]
            self.centroids = hsgr.gradients[:, 2:4]

    def update_centroids_and_gradients(self):
        """Update centroids and gradients instance variables.

        :Requires: ``hsgradients``
        :Updates: ``centroids``, ``gradients``

        """
        hsgr = self.hsgradients
        self.gradients = hsgr.gradients[:, 0:2]
        self.centroids = hsgr.gradients[:, 2:4]

    def determine_no_of_bases(self, order):
        """Determine the number of basis functions to use given the value of ``order``.

        After ``no_of_bases`` is determined, this method also prepares the basis
        functions that are needed for the calculation of the coefficients.

        :param order: a value from 1 to 6
        :type order: *integer*
        :Uses: :func:`~HS_WFP.build_bases`
        :Updates: ``order``, ``no_of_bases``

        """
        if order > 7:
            raise Exception(
                "The parameter order must be an integer " + "between 0 and 6."
            )
        self.order = order
        self.no_of_bases = HS_WFP.no_of_bases_array[order - 1]
        self.build_bases()

    def build_XY(self):
        """Build ``X`` and ``Y`` arrays

        :Requires: ``centroids``, ``to_unitize``, ``X_bases``, ``Y_bases``,
                   (optional) ``unitization_length``
        :Updates: ``X``, ``Y``

        """
        ctr = self.centroids.copy()

        if self.to_unitize is True:
            ctr = ctr / self.unitization_length

        Xc = [fn(*(ctr[:, 0], ctr[:, 1])) for fn in self.X_bases]
        Yc = [fn(*(ctr[:, 0], ctr[:, 1])) for fn in self.Y_bases]

        self.X = array(Xc)
        self.Y = array(Yc)

    def build_XY_nu(self):
        """Build non-unitized versions of ``X`` and ``Y``, and return them as output.

        :Requries: ``centroids``
        :Returns: the calculated arrays
        :rtype: *ndarray*

        """
        ctr = self.centroids.copy()

        Xc = [fn(*(ctr[:, 0], ctr[:, 1])) for fn in self.X_bases]
        Yc = [fn(*(ctr[:, 0], ctr[:, 1])) for fn in self.Y_bases]

        X = array(Xc)
        Y = array(Yc)

        return X, Y

    def build_M_matrix(self):
        """Build ``M`` using ``X`` and ``Y``

        :Requires: ``X``, ``Y``
        :Updates: ``M``

        """
        self.M = dot(self.X, self.X.T) + dot(self.Y, self.Y.T)

    def compute_poly_coeffs(self):
        """Calculate and update the coefficients of the polynomial basis functions

        :Requries: ``gradients``, ``no_of_bases``, ``poly_bases_names``
        :Uses: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.build_M_matrix`,
               :func:`~HS_WFP.compute_SSR`, (optional) :func:`~HS_WFP.deunitize`
        :Updates: ``b``, ``poly_coeffs_array``, ``poly_coeffs``, ``SSR``,
                  (optional) ``poly_coeffs_u``, (optional) ``poly_coeffs_array_u``

        """
        self.build_XY()
        self.build_M_matrix()

        gr = -self.gradients

        self.b = dot(self.X, gr[:, 0]) + dot(self.Y, gr[:, 1])

        if self.to_unitize is True:
            self.poly_coeffs_array_u = dot(linalg.inv(self.M), self.b)
            self.poly_coeffs_u = {
                self.poly_bases_names[i]: self.poly_coeffs_array_u[i]
                for i in arange(self.no_of_bases)
            }
            self.deunitize()
        else:
            self.poly_coeffs_array = dot(linalg.inv(self.M), self.b)

        self.poly_coeffs = {
            self.poly_bases_names[i]: self.poly_coeffs_array[i]
            for i in arange(self.no_of_bases)
        }

        self.compute_SSR()

    def compute_poly_coeffs_weighted(self, w):
        """Calculate and update the coefficients of the polynomial basis functions
        using the weighted ``gradients``, ``X`` and ``Y``.

        :param w: the weights to be used in the calculation
        :type w: *ndarray*
        :Requires: ``gradients``, ``poly_bases_names``, ``no_of_bases``
        :Uses: :func:`~HS_WFP.build_XY`, :func:`~HS_WFP.build_M_matrix`,
               :func:`~HS_WFP.compute_SSR`, (optional) :func:`~HS_WFP.deunitize`
        :Updates: ``weights``, ``b``, ``X``, ``Y``,
                  ``poly_coeffs_array``, ``poly_coeffs``, ``SSR``,
                  (optional) ``poly_coeffs_u``, (optional) ``poly_coeffs_array_u``

        """
        self.build_XY()
        if shape(w) != shape(self.centroids):
            raise Exception(
                "The shape of the weight parameter must be "
                + "the same as the shape of the centroids array."
            )

        self.weights = w
        wx = tile(w[:, 0], (shape(self.X)[0], 1))
        wy = tile(w[:, 1], (shape(self.X)[0], 1))
        self.X = self.X / wx
        self.Y = self.Y / wy
        wgr = -self.gradients / w
        self.build_M_matrix()

        self.b = dot(self.X, wgr[:, 0]) + dot(self.Y, wgr[:, 1])

        if self.to_unitize is True:
            self.poly_coeffs_array_u = dot(linalg.inv(self.M), self.b)
            self.poly_coeffs_u = {
                self.poly_bases_names[i]: self.poly_coeffs_array_u[i]
                for i in arange(self.no_of_bases)
            }
            self.deunitize()
        else:
            self.poly_coeffs_array = dot(linalg.inv(self.M), self.b)

        self.poly_coeffs = {
            self.poly_bases_names[i]: self.poly_coeffs_array[i]
            for i in arange(self.no_of_bases)
        }

        self.compute_SSR()

    def compute_SSR(self):
        """Calcualte the squared sum of the residues between the gradients used in the
        fit and the gradients synthesized using the coefficients obtained from the fit

        :Requires: ``gradients``
        :Uses: :func:`~HS_WFP.calculate_gradients`
        :Updates: ``SSR``

        """
        gr = self.calculate_gradients()
        self.SSR = ((self.gradients - gr) ** 2).sum()

    def calculate_gradients(self, coeffs=None):
        """Calculate the gradients from the coefficients of the polynomial basis functions

        If the input parameter ``coeffs`` is not given, the existing instance variable
        of the coefficients will be used.

        If the input parameter ``coeffs`` is given, its length must be equal to the number
        of bases ``no_of_bases``.

        :param coeffs: an array of the coefficients for the polynomial basis functions
        :type coeffs: *ndarray*
        :Requires: (optional) ``poly_coeffs_array``
        :Uses: (optional) :func:`~HS_WFP.build_XY_nu`
        :Returns: the gradients calculated from the coefficients
        :rtype: *ndarray*

        """
        # if coeffs is None:
        #     if self.to_unitize is True:
        #         coeffs = self.poly_coeffs_array_u
        #     else:
        #         coeffs = self.poly_coeffs_array

        if coeffs is None:
            coeffs = self.poly_coeffs_array

        if self.to_unitize is True:
            X, Y = self.build_XY_nu()
        else:
            X = self.X
            Y = self.Y

        grx = -1 * (tile(coeffs, (size(X[0]), 1)).T * X).sum(0)
        gry = -1 * (tile(coeffs, (size(Y[0]), 1)).T * Y).sum(0)
        gr = array([grx, gry]).T

        return gr

    def calculate_wf(
        self,
        coeffs=None,
        resolution=(1024, 1024),
        origin=(511.5, 511.5),
        spacing=(12e-6, 12e-6),
    ):
        """Generate a wavefront from the polynomial coefficients.

        :param coeffs: the coefficients for the polynomial basis functions
        :type coeffs: *ndarray*
        :param resolution: the resolution of the wavefront to be generated
        :type resolution: *tuple*
        :param origin: the coordinates of the origin in pixel units
        :type origin: *tuple*
        :param spacing: the vertical and horizontal spacings between the grid points
        :type spacing: *tuple*
        :Requires: ``poly_all_bases``, (optional) ``poly_coeffs_array``
        :Returns: a two-dimentional array of the generated wavefront
        :rtype: *ndarray*

        """

        xr = (arange(0, resolution[0]) - origin[0]) * spacing[0]
        yr = (arange(0, resolution[1]) - origin[1]) * spacing[1]

        xg, yg = meshgrid(xr, yr)
        wf = zeros(resolution)

        if coeffs is None:
            coeffs = self.poly_coeffs_array

        for k, pc in enumerate(coeffs):
            wf = wf + coeffs[k] * self.poly_all_bases[k](*(xg, yg))

        return wf - wf.min()

    def deunitize(self):
        """Rescale the coefficients of the unitized basis functions back to the values in
        meter units.

        :Requires: ``unitization_length``, ``poly_coeffs_array_u``,
                   ``poly_bases_names``, ``no_of_bases``
        :Updates: ``poly_coeffs_array``, ``poly_coeffs``

        """
        ul = self.unitization_length
        duarr_all = array(
            [
                1,
                1,
                ul,
                ul,
                ul,
                ul ** 2,
                ul ** 2,
                ul ** 2,
                ul ** 2,
                ul ** 3,
                ul ** 3,
                ul ** 3,
                ul ** 3,
                ul ** 3,
                ul ** 4,
                ul ** 4,
                ul ** 4,
                ul ** 4,
                ul ** 4,
                ul ** 4,
                ul ** 5,
                ul ** 5,
                ul ** 5,
                ul ** 5,
                ul ** 5,
                ul ** 5,
                ul ** 5,
            ]
        )
        duarr = duarr_all[0 : self.no_of_bases]

        self.poly_coeffs_array = self.poly_coeffs_array_u / duarr

        self.poly_coeffs = {
            self.poly_bases_names[i]: self.poly_coeffs_array[i]
            for i in arange(self.no_of_bases)
        }

    def compute_seidel_coeffs(self, coeffs=None):
        """Compute the coefficients of some Seidel functions from the polynomial
        coefficients.

        If the input parameter ``coeffs`` is given, this method returns the Seidel
        coefficients calculated from them without updating the instance variable
        ``seidel_coeffs``..

        If ``coeffs`` is not given, this method calculates the Seidel coefficients
        from the instance variable ``poly_coeffs_array`` and updates ``seidel_coeffs``.

        This method expects ``coeffs`` to be an array of at least 5 elements, or a
        *dict* containing at least the following fields: 'x', 'y', 'xy', 'x2' and
        'y2'.

        :param coeffs: the coefficients of the polynomials
        :type coeffs: *ndarray* or *dict*
        :Requires: (optional) ``poly_coeffs_array``
        :Updates: (optional) ``seidel_coeffs``
        :Returns: (optional) the Seidel coefficients
        :rtype: *dict*

        """
        seidel_coeffs = {}
        update_var = False

        if coeffs is None:
            coeffs_arr = self.poly_coeffs_array.copy()
            update_var = True
        elif type(coeffs) is dict:
            coeffs_arr = [
                coeffs["x"],
                coeffs["y"],
                coeffs["x2"],
                coeffs["xy"],
                coeffs["y2"],
            ]
        elif (type(coeffs) is not list) or (type(coeffs) is not ndarray):
            raise Exception(
                "The input parameter coeffs must be a dict, "
                + "a list or a numpy array of the numeric values."
            )
        else:
            coeffs_arr = coeffs

        P = coeffs_arr[2]
        Q = coeffs_arr[3]
        R = coeffs_arr[4]

        seidel_coeffs["prism_x"] = coeffs_arr[0]
        seidel_coeffs["prism_y"] = coeffs_arr[1]
        seidel_coeffs["prism"] = (coeffs_arr[0] ** 2 + coeffs_arr[1] ** 2) ** 0.5
        seidel_coeffs["alpha"] = arctan2(coeffs_arr[1], coeffs_arr[0])
        seidel_coeffs["cylindrical_power"] = 2 * ((R - P) ** 2 + Q ** 2) ** 0.5
        seidel_coeffs["spherical_power"] = (
            P + R - 0.5 * (seidel_coeffs["cylindrical_power"])
        )
        seidel_coeffs["phi"] = 0.5 * arctan2(-Q, R - P)

        if update_var is True:
            self.seidel_coeffs = seidel_coeffs
        else:
            return seidel_coeffs

    def compute_zernike_coeffs(self, coeffs=None):
        """Compute the coefficients of some Zernike functions from the polynomial
        coefficients.

        If the input parameter ``coeffs`` is given, this method returns the Zernike
        coefficients calculated from them without updating the instance variable
        ``seidel_coeffs``..

        If ``coeffs`` is not given, this method calculates the Zernike coefficients
        from the instance variable ``poly_coeffs_array`` and updates ``seidel_coeffs``.

        This method expects ``coeffs`` to be an array of at least 5 elements, or a
        *dict* containing at least the following fields: 'x', 'y', 'xy', 'x2' and
        'y2'.

        Note that the Zernike coefficients calculated with this method are not those
        of true Zernike functions, as they are not scaled to be orthogonal over
        a disc of size comparable to the image size. The method will need to be
        revised to apply appropriate scaling to the Zernike functions.

        :param coeffs: the coefficients of the polynomials
        :type coeffs: *ndarray* or *dict*
        :Requires: (optional) ``poly_coeffs_array``
        :Updates: (optional) ``seidel_coeffs``
        :Returns: (optional) the Seidel coefficients
        :rtype: *dict*

        """
        update_var = False

        if coeffs is None:
            update_var = True
            coeffs_arr = self.poly_coeffs_array
        elif type(coeffs) is dict:
            coeffs_arr = [
                coeffs["x"],
                coeffs["y"],
                coeffs["x2"],
                coeffs["xy"],
                coeffs["y2"],
            ]
        elif (type(coeffs) is not list) or (type(coeffs) is not ndarray):
            raise Exception(
                "The input parameter coeffs must be a dict, "
                + "a list or a numpy array of the numeric values."
            )
        else:
            coeffs_arr = coeffs

        z = zeros(5)
        z[0] = coeffs_arr[0]
        z[1] = coeffs_arr[1]
        z[2] = (coeffs_arr[2] + coeffs_arr[4]) / 4.0
        z[3] = (coeffs_arr[2] - coeffs_arr[4]) / 2.0
        z[4] = coeffs_arr[3] / 2.0

        if update_var is True:
            self.zernike_coeffs = z

    def build_bases(self):
        """Set up the basis functions according to ``no_of_bases``.

        This method simply takes a number of the polynomial functions from ``X_all_bases``
        ``Y_all_bases`` and the polynomial basis function names from
        ``poly_all_bases_names``. The number of items to take is determined by
        ``no_of_bases``.

        :Requires: ``X_all_bases``, ``Y_all_bases``, ``poly_all_bases_names``,
                   ``no_of_bases``
        :Updates: ``X_bases``, ``Y_bases``, ``poly_bases_names``

        """
        self.X_bases = self.X_all_bases[0 : (self.no_of_bases)]
        self.Y_bases = self.Y_all_bases[0 : (self.no_of_bases)]
        self.poly_bases_names = self.poly_all_bases_names[0 : (self.no_of_bases)]

    def build_all_bases(self):
        """Set up the X, Y and polynomial basis functions and the names of the bases
        functions that are available in this class.

        This method prepares sets of all basis functions that can be used in the least
        squares fit of gradients and the construction of wavefronts.

        :func:`~HS_WFP.build_bases` can then take subsets of the functions from these
        sets. ``no_of_bases``, derived from ``order``, determines how many functions
        should be taken.

        :Updates: ``X_all_bases``, ``Y_all_bases``, ``poly_all_bases``,
                  ``poly_all_bases_names``

        """
        self.X_all_bases = [
            lambda x, y: ones(size(x)),
            lambda x, y: zeros(size(x)),
            lambda x, y: 2 * x,
            lambda x, y: y,
            lambda x, y: zeros(size(x)),
            lambda x, y: 3 * x ** 2,
            lambda x, y: 2 * x * y,
            lambda x, y: y ** 2,
            lambda x, y: zeros(size(x)),
            lambda x, y: 4 * x ** 3,
            lambda x, y: 3 * x ** 2.0 * y,
            lambda x, y: 2 * x * y ** 2,
            lambda x, y: y ** 3,
            lambda x, y: zeros(size(x)),
            lambda x, y: 5 * x ** 4,
            lambda x, y: 4 * x ** 3 * y,
            lambda x, y: 3 * x ** 2 * y ** 2,
            lambda x, y: 2 * x * y ** 3,
            lambda x, y: y ** 4,
            lambda x, y: zeros(size(x)),
            lambda x, y: 6 * x ** 5,
            lambda x, y: 5 * x ** 4.0 * y,
            lambda x, y: 4 * x ** 3.0 * y ** 2,
            lambda x, y: 3 * x ** 2 * y ** 3,
            lambda x, y: 2 * x * y ** 4,
            lambda x, y: y ** 5,
            lambda x, y: zeros(size(x)),
        ]

        self.Y_all_bases = [
            lambda x, y: zeros(size(x)),
            lambda x, y: ones(size(x)),
            lambda x, y: zeros(size(x)),
            lambda x, y: x,
            lambda x, y: 2 * y,
            lambda x, y: zeros(size(x)),
            lambda x, y: x ** 2,
            lambda x, y: 2 * x * y,
            lambda x, y: 3 * y ** 2,
            lambda x, y: zeros(size(x)),
            lambda x, y: x ** 3,
            lambda x, y: 2 * x ** 2 * y,
            lambda x, y: 3 * x * y ** 2,
            lambda x, y: 4 * y ** 3,
            lambda x, y: zeros(size(x)),
            lambda x, y: x ** 4,
            lambda x, y: 2 * x ** 3 * y,
            lambda x, y: 3 * x ** 2 * y ** 2,
            lambda x, y: 4 * x * y ** 3,
            lambda x, y: 5 * y ** 4,
            lambda x, y: zeros(size(x)),
            lambda x, y: x ** 5,
            lambda x, y: 2 * x ** 4 * y,
            lambda x, y: 3 * x ** 3 * y ** 2,
            lambda x, y: 4 * x ** 2 * y ** 3,
            lambda x, y: 5 * x * y ** 4,
            lambda x, y: 6 * y ** 5,
        ]

        self.poly_all_bases = [
            lambda x, y: x,
            lambda x, y: y,
            lambda x, y: x ** 2,
            lambda x, y: x * y,
            lambda x, y: y ** 2,
            lambda x, y: x ** 3,
            lambda x, y: x ** 2 * y,
            lambda x, y: x * y ** 2,
            lambda x, y: y ** 3,
            lambda x, y: x ** 4,
            lambda x, y: x ** 3 * y,
            lambda x, y: x ** 2 * y ** 2,
            lambda x, y: x * y ** 3,
            lambda x, y: y ** 4,
            lambda x, y: x ** 5,
            lambda x, y: x ** 4 * y,
            lambda x, y: x ** 3 * y ** 2,
            lambda x, y: x ** 2 * y ** 3,
            lambda x, y: x * y ** 4,
            lambda x, y: y ** 5,
            lambda x, y: x ** 6,
            lambda x, y: x ** 5 * y,
            lambda x, y: x ** 4 * y ** 2,
            lambda x, y: x ** 3 * y ** 3,
            lambda x, y: x ** 2 * y ** 4,
            lambda x, y: x * y ** 5,
            lambda x, y: y ** 6,
        ]

        self.poly_all_bases_names = [
            "x",
            "y",
            "x2",
            "xy",
            "y2",
            "x3",
            "x2y",
            "xy2",
            "y3",
            "x4",
            "x3y",
            "x2y2",
            "xy3",
            "y4",
            "x5",
            "x4y",
            "x3y2",
            "x2y3",
            "xy4",
            "y5",
            "x6",
            "x5y",
            "x4y2",
            "x3y3",
            "x2y4",
            "xy5",
            "y6",
        ]



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""
from numpy import *
import scipy.ndimage
import matplotlib.pyplot as plt
from copy import deepcopy


class HS_Gradients:
    """A class to construct the gradients from the two sets of centroids.

    ``HS_Gradients`` is a class reponsible for construction and accees of the
    gradients. The gradient fields are constructed by calculating the displacements
    between two sets of centroids, and then scaling the result with the other
    parameters like ``pixel_size``, ``lever_arm`` and ``magnification``.

    **Instantiation**

    The class accepts three input parameters for the instantiation. ``hsc_i``, and
    ``hsc_f`` are initial (or reference) and final (or live) instances of
    :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>`.

    ``lever_arm`` is the camera lever arm length, i.e., the distance between the
    CCD and the Hartmann plate in front of it.

    **Instance Variables**

    - ``pixel_size``: the width of a pixel (in meters), assuming each pixel is a square
      shape

      :Type: positive *float*
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``lever_arm``: the distance (in meters) between the CCD and the Hartmann plate

      :Type: positive *float*
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``magnification``: the magnification factor required to scale the gradients

      (may need more detailed description)

      :Type: *float*
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``origin``: the origin of the coordinates used for the locations of the gradient
      vectors

      The choice of origin affects the wavefront construction, and is normally set at
      the centre of the image.

      :Type: *ndarray* of two elements
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``initial_centroids``: an instance of :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>` that contains the initial (or reference) centroids

      :Type: instance of :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>`
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``final_centroids``: an instance of :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>` that contains the initial (or reference) centroids

      :Type: instance of :class:`HS_Centroids <HWS.HS_Centroids.HS_Centroids>`
      :Required by: :func:`~HS_Gradients.construct_gradients`

    - ``gradients``: An array of the gradients constructed from the initial and the
      final centroids

      Its shape is N by 4 (N = the number of centroids). The first two elements in
      each row are x and y values of the gradients (in radian, i.e., dimensionless
      units). The thrid and fourth elements are the coordinates of the corresponding
      reference centroid (meter units).

      :Type: *ndarray*
      :Required by: :func:`~HS_Gradients.calculate_rms`, :func:`~HS_Gradients.plot`
      :Updated by: :func:`~HS_Gradients.construct_gradients`

    - ``rms``: the root-mean-squqre values of the gradient magnitude values

      :Type: *ndarray* of N elements, where N is the number of centroids
      :Updated by: :func:`~HS_Gradients.calculate_rms`

    .. note:: The following two instance variables are not used but exist just
       to remind a user which units are used to express the centroids and gradients.

    - ``centroidUnits``
    - ``gradientUnits``

    """

    def __init__(self, hsc_i=None, hsc_f=None, lever_arm=0.01):
        self.pixel_size = 12e-6
        self.lever_arm = lever_arm
        self.magnification = 1.0
        self.origin = array([511.5, 511.5])
        self.initial_centroids = None
        self.final_centroids = None
        self.initial_centroids_input = hsc_i
        self.final_centroids_input = hsc_f
        self.gradients = None
        self.magnitudes = None
        self.rms = None
        self.centroidUnits = "( x [pixels], y [pixels] )"
        self.gradientUnits = "( dx/dL [rad], dy/dL [rad], x [meters], y [meters] )"

        if (hsc_i is not None) and (hsc_f is not None):
            self.remove_invalid_centroids()
        else:
            # just in case a user assigns one of the instances...
            self.initial_centroids = hsc_i
            self.final_centroids = hsc_f

    def construct_gradients(self):
        """Construct gradients from ``initial_centroids`` and ``final_centroids``.

        :Requires: ``pixel_size``, ``lever_arm``, ``magnification``, ``origin``,
                   ``initial_centroids``, ``final_centroids``
        :Updates: ``gradients``

        """
        self.lever_arm = float(self.lever_arm)
        self.magnigication = float(self.magnification)
        self.origin = self.origin.astype(float)

        x_i = self.initial_centroids.centroids[:, 0] - self.origin[0]
        y_i = self.initial_centroids.centroids[:, 1] - self.origin[1]

        x_i = x_i * (self.pixel_size * self.magnification)
        y_i = y_i * (self.pixel_size * self.magnification)

        displacements = (
            self.final_centroids.centroids - self.initial_centroids.centroids
        )

        mod_factor = self.pixel_size / (self.lever_arm * self.magnification)
        self.gradients = hstack((displacements * mod_factor, array([x_i, y_i]).T))
        # self.gradients = displacements
        self.magnitudes = sqrt(self.gradients[:, 0] ** 2 + self.gradients[:, 1] ** 2)

    def remove_invalid_centroids(self):
        """Removes the rows containing None from input centroids.

        (Experimental)
        """
        c_i = self.initial_centroids_input.centroids
        c_f = self.final_centroids_input.centroids
        indices_to_remove = unique(argwhere(equal(c_f, None))[:, 0])

        self.initial_centroids = deepcopy(self.initial_centroids_input)
        self.final_centroids = deepcopy(self.final_centroids_input)

        self.initial_centroids.hsimage = self.initial_centroids_input.hsimage
        self.final_centroids.hsimage = self.final_centroids_input.hsimage

        c_i_r = delete(c_i, indices_to_remove, axis=0)
        c_f_r = delete(c_f, indices_to_remove, axis=0)

        self.initial_centroids.centroids = c_i_r.astype(float)
        self.final_centroids.centroids = c_f_r.astype(float)

        self.initial_centroids.no_of_centroids = shape(
            self.initial_centroids.centroids
        )[0]

        self.final_centroids.no_of_centroids = shape(self.final_centroids.centroids)[0]

        if self.initial_centroids.intensities is not None:
            self.initial_centroids.intensities = delete(
                self.initial_centroids.intensities, indices_to_remove, axis=0
            )

        if self.final_centroids.intensities is not None:
            self.final_centroids.intensities = delete(
                self.final_centroids.intensities, indices_to_remove, axis=0
            )

    def calculate_rms(self):
        """Calculate root-mean-square of gradient magnitudes.

        :Requires: ``gradients``
        :Updates: ``rms``
        :Returns: the calcualted rms value
        :rtype: *float*

        """
        noc = float(shape(self.gradients)[0])
        self.rms = (((self.gradients[:, 0:2]) ** 2).sum() / noc) ** 0.5
        return self.rms

    def plot(self):
        """Plot the gradients using quiver plot method of matplotlib module

        :Requires: ``gradients``
        """
        plt.quiver(
            self.gradients[:, 2],
            self.gradients[:, 3],
            self.gradients[:, 0],
            self.gradients[:, 1],
        )
        plt.gca().set_aspect("equal", adjustable="box")
        plt.draw()
        # plt.show(block=False)
        plt.show()


# Won Kim 26/11/2018 3:19:31 PM



"""
Copyright (C) University of Adelaide, Australia - All Rights Reserved
This source code is protected under proprietary reference only license.
Please refer to the accompanying LICENSE file for further information.
"""

from numpy import *


class HS_BeamProfile:
    def __init__(self):

        self.hsgradients = None
        self.gradients = None
        self.centroids = None
        self.origin = (511.5, 511.5)
        self.intensities = None
        self.grid_spacing = 12e-6
        self.spot_spacing = None
        self.first_moments = {}
        self.second_moments = {}
        self.intensity_gradients = None
        self.P_matrix = None
        self.width_x = None
        self.width_y = None
        self.div_x = None
        self.div_y = None
        self.roc_x = None
        self.roc_y = None
        self.M2_x = None
        self.M2_y = None
        self.wavelength = None

    def convert_centroids(self):
        centroids = self.hsgradients.initial_centroids.centroids
        c_m = zeros(shape(centroids))
        c_m[:, 0] = centroids[:, 0] - self.origin[0]
        c_m[:, 1] = centroids[:, 1] - self.origin[1]
        self.centroids = c_m * self.grid_spacing

    def construct_beam_profile(self):
        # self.construct_intensity_gradients;
        self.construct_first_moments()
        self.construct_second_moments()
        # self.construct_P;
        self.width_x = 4 * sqrt(self.second_moments["x2"])
        self.width_y = 4 * sqrt(self.second_moments["y2"])
        # self.div_x = 4*sqrt(self.second_moments.u2);
        # self.div_y = 4*sqrt(self.second_moments.v2);
        self.roc_x = self.second_moments["x2"] / self.second_moments["xu"]
        self.roc_y = self.second_moments["y2"] / self.second_moments["yv"]
        # self.M2_x = (4*pi/self.wavelength)* ...
        #            sqrt(self.second_moments.x2 * ...
        #                 self.second_moments.u2 - ...
        #                 self.second_moments.xu^2);
        # self.M2_y = (4*pi/self.wavelength)* ...
        #            sqrt(self.second_moments.y2 * ...
        #                 self.second_moments.v2 - ...
        #                 self.second_moments.yv^2);

    def construct_first_moments(self):
        if (self.centroids is None) or (self.intensities is None):
            raise Exception(
                "Instance variables centroids and intensities "
                + "must be present for the moments calculation."
            )
        totI = self.intensities.sum()
        avC = dot(self.centroids.T, self.intensities) / totI
        self.first_moments["x"] = avC[0]
        self.first_moments["y"] = avC[1]

        gr = self.gradients[:, 0:2]
        avG = dot(gr.T, self.intensities) / totI
        self.first_moments["u"] = avG[0]
        self.first_moments["v"] = avG[1]

    def construct_second_moments(self):
        self.construct_x2()
        self.construct_xy()
        self.construct_xu()
        self.construct_xv()
        self.construct_y2()
        self.construct_yu()
        self.construct_yv()
        self.construct_uv()

    def construct_x2(self):
        self.second_moments["x2"] = (
            (self.centroids[:, 0] - self.first_moments["x"]) ** 2 * self.intensities
        ).sum() / self.intensities.sum()

    def construct_xy(self):
        self.second_moments["xy"] = (
            (self.centroids[:, 0] - self.first_moments["x"])
            * (self.centroids[:, 1] - self.first_moments["y"])
            * self.intensities
        ).sum() / self.intensities.sum()

    def construct_xu(self):
        self.second_moments["xu"] = (
            (self.centroids[:, 0] - self.first_moments["x"])
            * (self.gradients[:, 0] - self.first_moments["u"])
            * self.intensities
        ).sum() / self.intensities.sum()

    def construct_xv(self):
        self.second_moments["xv"] = (
            (self.centroids[:, 0] - self.first_moments["x"])
            * (self.gradients[:, 1] - self.first_moments["v"])
            * self.intensities
        ).sum() / self.intensities.sum()

    def construct_y2(self):
        self.second_moments["y2"] = (
            (self.centroids[:, 1] - self.first_moments["y"]) ** 2 * self.intensities
        ).sum() / self.intensities.sum()

    def construct_yu(self):
        self.second_moments["yu"] = (
            (self.centroids[:, 1] - self.first_moments["y"])
            * (self.gradients[:, 0] - self.first_moments["u"])
            * self.intensities
        ).sum() / self.intensities.sum()

    def construct_yv(self):
        self.second_moments["yv"] = (
            (self.centroids[:, 1] - self.first_moments["y"])
            * (self.gradients[:, 1] - self.first_moments["v"])
            * self.intensities
        ).sum() / self.intensities.sum()

    def construct_uv(self):
        self.second_moments["uv"] = (
            (self.gradients[:, 0] - self.first_moments["u"])
            * (self.gradients[:, 1] - self.first_moments["v"])
            * self.intensities
        ).sum() / self.intensities.sum()
